"""Type definitions for the GA and GS IDs."""

from functools import partial
from typing import Annotated, Literal

from pydantic import AfterValidator

from gso.services.subscriptions import is_resource_type_value_unique


def validate_id(value: str, prefix: Literal["GA", "GS"], field_name: str) -> str:
    """Validate that the ID is unique, has the correct prefix, and is within valid constraints.

    Args:
        value: The ID value to validate.
        prefix: The required prefix for the ID.
        field_name: The database field name to check for uniqueness.

    Raises:
        ValueError: If the ID is not valid.

    Returns:
        str: The validated ID.
    """
    min_range = 50000
    max_range = 99999
    if not value.startswith(prefix):
        err = f"{field_name} must start with the prefix '{prefix}'."
        raise ValueError(err)

    try:
        numeric_part = int(value.split("-")[-1])
    except ValueError:
        err = f"{field_name} must have a numeric part after the prefix '{prefix}'."
        raise ValueError(err) from ValueError

    if min_range <= numeric_part <= max_range:
        err = f"{field_name} must not have a numeric part between {min_range} and {max_range}, received {numeric_part}"
        raise ValueError(err)

    if not is_resource_type_value_unique(field_name, value):
        err = f"{field_name} must be unique, {value} is already in use."
        raise ValueError(err)

    return value


IMPORTED_GA_ID = Annotated[str, AfterValidator(partial(validate_id, prefix="GA", field_name="ga_id"))]
IMPORTED_GS_ID = Annotated[str, AfterValidator(partial(validate_id, prefix="GS", field_name="gs_id"))]
