"""A creation workflow that adds an existing Edge Port to the DB."""

from typing import Annotated, Any
from uuid import uuid4

from orchestrator import workflow
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle
from orchestrator.workflow import StepList, begin, done, step
from orchestrator.workflows.steps import resync, set_status, store_process_subscription
from pydantic import AfterValidator, ConfigDict
from pydantic_forms.types import UUIDstr
from pydantic_forms.validators import validate_unique_list

from gso.products import ProductName
from gso.products.product_blocks.edge_port import EdgePortAEMemberBlockInactive, EdgePortType, EncapsulationType
from gso.products.product_types.edge_port import ImportedEdgePortInactive
from gso.products.product_types.router import Router
from gso.services.partners import get_partner_by_name
from gso.services.subscriptions import get_product_id_by_name
from gso.utils.helpers import active_pe_router_selector
from gso.utils.types.geant_ids import IMPORTED_GA_ID
from gso.utils.types.interfaces import LAGMember, PhysicalPortCapacity


@step("Create subscription")
def create_subscription(partner: str) -> State:
    """Create a new subscription object."""
    partner_id = get_partner_by_name(partner)["partner_id"]
    product_id = get_product_id_by_name(ProductName.IMPORTED_EDGE_PORT)
    subscription = ImportedEdgePortInactive.from_product_id(product_id, partner_id)

    return {"subscription": subscription, "subscription_id": subscription.subscription_id}


def initial_input_form_generator() -> FormGenerator:
    """Generate a form that is filled in using information passed through the API endpoint."""

    class ImportEdgePort(FormPage):
        model_config = ConfigDict(title="Import Edge Port")

        node: active_pe_router_selector()  # type: ignore[valid-type]
        partner: str
        service_type: EdgePortType
        enable_lacp: bool
        speed: PhysicalPortCapacity
        encapsulation: EncapsulationType = EncapsulationType.DOT1Q
        minimum_links: int
        mac_address: str | None = None
        ignore_if_down: bool = False
        ga_id: IMPORTED_GA_ID | None = None
        description: str | None = None
        name: str
        ae_members: Annotated[list[LAGMember], AfterValidator(validate_unique_list)]

    user_input = yield ImportEdgePort

    return user_input.model_dump()


@step("Initialize subscription")
def initialize_subscription(
    subscription: ImportedEdgePortInactive,
    node: UUIDstr,
    service_type: EdgePortType,
    speed: PhysicalPortCapacity,
    encapsulation: EncapsulationType,
    name: str,
    minimum_links: int,
    ga_id: IMPORTED_GA_ID | None,
    mac_address: str | None,
    partner: str,
    enable_lacp: bool,  # noqa: FBT001
    ignore_if_down: bool,  # noqa: FBT001
    ae_members: list[dict[str, Any]],
    description: str | None = None,
) -> State:
    """Initialise the subscription object in the service database."""
    router = Router.from_subscription(node).router
    subscription.edge_port.node = router
    subscription.edge_port.edge_port_name = name
    subscription.edge_port.edge_port_description = description
    subscription.edge_port.enable_lacp = enable_lacp
    subscription.edge_port.encapsulation = encapsulation
    subscription.edge_port.mac_address = mac_address
    subscription.edge_port.member_speed = speed
    subscription.edge_port.minimum_links = minimum_links
    subscription.edge_port.edge_port_type = service_type
    subscription.edge_port.ignore_if_down = ignore_if_down
    subscription.edge_port.ga_id = ga_id
    subscription.description = f"Edge Port {name} on {router.router_fqdn}, {partner}, {ga_id or ""}"
    for member in ae_members:
        subscription.edge_port.edge_port_ae_members.append(
            EdgePortAEMemberBlockInactive.new(subscription_id=uuid4(), **member)
        )

    return {"subscription": subscription}


@workflow(
    "Import Edge Port",
    initial_input_form=initial_input_form_generator,
    target=Target.CREATE,
)
def create_imported_edge_port() -> StepList:
    """Import an Edge Port without provisioning it."""
    return (
        begin
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
