"""Workflow for terminating a LAN Switch interconnect."""

from orchestrator import begin, workflow
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import StepList, done, step
from orchestrator.workflows.steps import resync, set_status, store_process_subscription, unsync
from orchestrator.workflows.utils import wrap_modify_initial_input_form
from pydantic_forms.types import FormGenerator
from pydantic_forms.validators import Label

from gso.products.product_types.lan_switch_interconnect import LanSwitchInterconnect
from gso.services.infoblox import delete_host_by_fqdn, delete_network
from gso.utils.helpers import generate_lan_switch_interconnect_subnet_v4
from gso.utils.types.tt_number import TTNumber


def _input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    lan_switch_interconnect = LanSwitchInterconnect.from_subscription(subscription_id)

    class TerminateForm(FormPage):
        if lan_switch_interconnect.status == SubscriptionLifecycle.INITIAL:
            info_label: Label = (
                "This will immediately mark the subscription as terminated, preventing any other workflows from "
                "interacting with this product subscription."
            )
            info_label_2: Label = "ONLY EXECUTE THIS WORKFLOW WHEN YOU ARE ABSOLUTELY SURE WHAT YOU ARE DOING."

        tt_number: TTNumber

    yield TerminateForm
    return {"subscription": lan_switch_interconnect}


@step("Release IPAM resources")
def clean_up_ipam(subscription: LanSwitchInterconnect) -> None:
    """Clear up :term:`IPAM` resources in :term:`DNS`.

    First, remove :term:`DNS` records for both the router and switch side of the interconnect.
    Second, remove the registered network for this interconnect.
    """
    delete_host_by_fqdn(subscription.lan_switch_interconnect.switch_side.switch.fqdn)
    delete_host_by_fqdn(
        f"{subscription.lan_switch_interconnect.router_side.ae_iface}."
        f"{subscription.lan_switch_interconnect.router_side.node.router_fqdn}"
    )
    delete_network(
        generate_lan_switch_interconnect_subnet_v4(
            subscription.lan_switch_interconnect.router_side.node.router_site.site_internal_id
        )
    )


@workflow(
    "Terminate LAN Switch Interconnect",
    initial_input_form=wrap_modify_initial_input_form(_input_form_generator),
    target=Target.TERMINATE,
)
def terminate_lan_switch_interconnect() -> StepList:
    """Terminate a LAN Switch Interconnect."""
    return (
        begin
        >> store_process_subscription(Target.TERMINATE)
        >> unsync
        >> clean_up_ipam
        >> set_status(SubscriptionLifecycle.TERMINATED)
        >> resync
        >> done
    )
