"""A creation workflow that adds an existing opengear to the service database."""

from orchestrator import workflow
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle
from orchestrator.workflow import StepList, done, init, step
from orchestrator.workflows.steps import resync, set_status, store_process_subscription
from pydantic import ConfigDict

from gso.products import ProductName
from gso.products.product_types.opengear import ImportedOpengearInactive
from gso.services.partners import get_partner_by_name
from gso.services.subscriptions import get_product_id_by_name, get_site_by_name
from gso.utils.types.ip_address import IPv4AddressType


@step("Create subscription")
def create_subscription(partner: str) -> State:
    """Create a new subscription object."""
    partner_id = get_partner_by_name(partner)["partner_id"]
    product_id = get_product_id_by_name(ProductName.IMPORTED_OPENGEAR)
    subscription = ImportedOpengearInactive.from_product_id(product_id, partner_id)

    return {
        "subscription": subscription,
        "subscription_id": subscription.subscription_id,
    }


def initial_input_form_generator() -> FormGenerator:
    """Generate a form that is filled in using information passed through the API endpoint."""

    class ImportOpengear(FormPage):
        model_config = ConfigDict(title="Import Opengear")

        partner: str
        opengear_site: str
        opengear_hostname: str
        opengear_wan_address: IPv4AddressType
        opengear_wan_netmask: IPv4AddressType
        opengear_wan_gateway: IPv4AddressType

    user_input = yield ImportOpengear

    return user_input.model_dump()


@step("Initialize subscription")
def initialize_subscription(
    subscription: ImportedOpengearInactive,
    opengear_site: str,
    opengear_hostname: str,
    opengear_wan_address: IPv4AddressType | None,
    opengear_wan_netmask: IPv4AddressType | None,
    opengear_wan_gateway: IPv4AddressType | None,
) -> State:
    """Initialise the Imported Opengear subscription using input data."""
    subscription.opengear.opengear_site = get_site_by_name(opengear_site).site
    subscription.opengear.opengear_hostname = opengear_hostname
    subscription.opengear.opengear_wan_address = opengear_wan_address
    subscription.opengear.opengear_wan_netmask = opengear_wan_netmask
    subscription.opengear.opengear_wan_gateway = opengear_wan_gateway

    return {"subscription": subscription}


@workflow(
    "Import Opengear",
    initial_input_form=initial_input_form_generator,
    target=Target.CREATE,
)
def create_imported_opengear() -> StepList:
    """Import an Opengear without provisioning it."""
    return (
        init
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
